{{
      $Id$
   Author: Perry Harrington
Copyright: (c) 2012 Perry Harrington
=======================================================================

This is a demo program to show the usage of the SPI SPIN object.

This demo reads the current temperature from a Maxim DS1620 digital
thermometer and thermostat and displays the temperature on the Parallax
Serial Terminal.  You can obtain the datasheet from Maxim
Semiconductor.

This device is an 8 pin DIP with an active high CLK input and single DQ
data line instead of MISO and MOSI I/O.  The chip uses LSB first
endianness, data is read PRE clock. The maximum clock rate is 1.75Mhz.

To send a command to the chip you must set RST high, send the command, then
signal the end of command by setting RST low.  When receiving data set RST
low to signal the end of reception.

The output of the chip is a 9 bit signed temperature in 1/2 degree Celsius
units.  That means you must divide the result by 2 to get whole degrees.

}}
CON

_clkmode = xtal1 + pll16x
_xinfreq = 5_000_000

DPIN    = 0     'Data pin
CPIN    = 1     'Clock pin
RST     = 2     'Reset pin
CLKr    = 10    '100us clock frequency

START_CONVERT   = $EE           'Start temp conversion
READ_TEMP       = $AA           'Read temp from DS1620
WRITE_CONFIG    = $0C           'Write config register
CONFIG_REG      = %0000_1010    'enable continuous conversion and CPU 3wire
CONVERT_TIME    = 750           'Takes 750ms to initialize temp conversion
EEPROM_WTIME    = 10            'Takes 10ms to write to EEPROM config

OBJ

pst: "Parallax Serial Terminal"
spi: "SPI_spinv1.1"

VAR
  long temperature

PUB main
{{
  This function does the setup for the DS1620 and reads the current temperature once per
  second, converts it to Fahrenheit, and displays it on the serial terminal
}}

  pst.Start(115200)                                     'debug output
  pst.Str(String("Welcome to the SPI demo! This program prints the current"))
  pst.Newline
  pst.Str(String("temperature from a DS1620 digital thermometer chip."))
  pst.Newline
  pst.Str(String("Binary    Hex  Fahrenheit"))
  pst.Newline
  spi.Start(CLKr,spi#HIGH)                              'initial clock high, going low on first bit

  DIRA[RST]~~                                           'enable RST as output
  OUTA[RST]~                                            'set RST low

  Toggle(RST)                                           'clear the RST line so we can talk to chip

  spi.ShiftOut(DPIN,CPIN,spi#LSB,8,WRITE_CONFIG)        'send the config write command
  spi.ShiftOut(DPIN,CPIN,spi#LSB,8,CONFIG_REG)          'send config register contents

  Toggle(RST)

  PauseMS(EEPROM_WTIME)                                 'wait 10ms for config to write to EEPROM

  Toggle(RST)

  spi.SHIFTOUT(DPIN,CPIN,spi#LSBFIRST,8,START_CONVERT)  'send beginning of conversion command

  Toggle(RST)

  PauseMS(CONVERT_TIME)                                 'wait for first conversion

  repeat
    Toggle(RST)

    spi.ShiftOut(DPIN,CPIN,spi#LSB,8,READ_TEMP)         'send temp reading command
    temperature := spi.ShiftIn(DPIN,CPIN,spi#LSB+spi#PRE, 9)                   'fetch temperature value

    Toggle(RST)

    pst.Bin(temperature,9)                              'print out raw value binary
    pst.Char(32)
    pst.Hex(temperature,4)                              'print out raw hex value
    pst.Char(32)

    temperature := temperature << 23 ~> 23              'zero, sign extend, and convert to whole degrees
    temperature *=5

    temperature := temperature * 9 / 5 + 320            'convert reading to fahrenheit

    pst.Dec(temperature/10)                             'print out temp in whole degrees Fahrenheit
    pst.Char(".")
    pst.Dec(temperature // 10)                          'print out tenths using modulus operator
    pst.Char("°")
    pst.Newline

    PauseMS(1000)                                       'update once per second

PRI PauseMS(_time)
  waitcnt(CLKFREQ / 1000 * _time + cnt)

PRI Toggle(_pin)
  !OUTA[_pin]

DAT
{{
┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │                                                            
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │ 
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}
