%{

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

extern char *yytext;
extern int lineno;
extern int yylineno;
extern int yy_start;

#define YY_START (((yy_start) - 1) / 2)


int	if_level	= 0;
int	loop_level	= 0;
int	comment_type	= 0;

int block_type;

void yyerror(const char *msg)
{
//        va_list ap;

//        va_start(ap,msg);

        printf("error: %5d:%3d: %s\n", lineno, YY_START, msg);
//        vprintf(msg,ap);
//        printf("\n");

//        va_end(ap);

        //exit(1);
}

extern int nblock_types[];
char *tblock_types[] = { "CON", "OBJ", "VAR", "DAT", "PUB", "PRI" };

void print_block_type(void)
{
	char *type;
	int	i=0;

	do {
		if (nblock_types[i] == block_type) {
			type = tblock_types[i];
			break;
		}
		i++;
	} while (1);

	printf("Block type changed: %d: %s\n", block_type,type);
}

%}

%defines

%union {
	char	*string;
	float	f;
	long	l;
	int	operator;
}

%token	CON
%token	OBJ
%token	VAR
%token	DAT
%token	PUB
%token	PRI
%token	NEWLINE	
%token	BLOCK_COMMENT_BEGIN MLCODE_COMMENT_BEGIN
%token	BLOCK_COMMENT_CLOSE MLCODE_COMMENT_CLOSE
%token	BLOCK_COMMENT MLCODE_COMMENT
%token	DOC_COMMENT
%token	CODE_COMMENT
%token	REPEAT
%token	WHILE
%token	UNTIL
%token	FROM
%token	TO
%token	STEP
%token	CONTINUE
%token	QUIT
%token	IF
%token	IFNOT
%token	ELSE
%token	ELSEIF
%token	ELSEIFNOT
%token	CASE
%token	OTHER
%token	BOOL_OR_ASSIGN
%token	BOOL_OR
%token	BOOL_AND_ASSIGN
%token	BOOL_AND
%token	BOOL_NOT_ASSIGN
%token	BOOL_NOT
%token	ASSIGN
%token	LT
%token	GT
%token	NE NE_ASSIGN
%token	EQUAL		
%token	EQUAL_EQUAL EQUAL_EQUAL_ASSIGN
%token	EQUAL_LT EQUAL_LT_ASSIGN
%token	EQUAL_GT EQUAL_GT_ASSIGN
%token	LIMIT_MIN LIMIT_MIN_ASSIGN
%token	LIMIT_MAX LIMIT_MAX_ASSIGN
%token	ADD ADD_ASSIGN
%token	SUBTRACT SUBTRACT_ASSIGN
%token	MULTIPLY_LOW MULTIPLY_LOW_ASSIGN
%token	MULTIPLY_HIGH MULTIPLY_HIGH_ASSIGN
%token	DIVIDE DIVIDE_ASSIGN
%token	MODULUS MODULUS_ASSIGN
%token	OR OR_ASSIGN
%token	XOR XOR_ASSIGN
%token	AND AND_ASSIGN
%token	ROR ROR_ASSIGN
%token	ROL ROL_ASSIGN
%token	SHR SHR_ASSIGN
%token	SHL SHL_ASSIGN
%token	SAR SAR_ASSIGN
%token	REVERSE REVERSE_ASSIGN
%token	SQRT
%token	ABS
%token	DECODE
%token	ENCODE
%token	NOT
%token	DEC
%token	INC
%token	CLEAR
%token	SET
%token	RAND
%token	AT
%token	AT_AT
%right	EQUAL ASSIGN
%left	BOOL_OR	BOOL_AND BOOL_NOT	
%left	LT GT NE		
%left	EQUAL_EQUAL EQUAL_LT EQUAL_GT	
%left	LIMIT_MIN LIMIT_MAX	
%left	ADD SUBTRACT	
%left	MULTIPLY_LOW MULTIPLY_HIGH	
%left	DIVIDE	 MODULUS		
%left	OR XOR AND		
%left	ROR ROL SHR SHL SAR		
%left	REVERSE		
%left	SQRT ABS		
%left	DECODE ENCODE		
%left	NOT DEC INC		
%left	CLEAR SET		
%left	RAND		
%left	AT AT_AT
%token	<string> BYTE_T
%token	<string> WORD_T
%token	<string> LONG_T
%token	<string> TEXT WORD STRING
%token	<string> NUMBER FLOAT HEX BINARY
%type	<string> value constantvalue expression symbol
%type	<string> text_block mltext_block comment rvalue
%type	<string> parameters type var_decl dat_decl step

%%

statement_list:
	| statement_list statement
	;

statement:
	comment
	| con
	| obj
	| var
	| function_decl
	| dat
	| control_statement
	| expression
	| obj_decl
	| var_decl
	| dat_decl
	;

con:	CON	{
			block_type = CON;
			print_block_type();
		}

obj:	OBJ	{
			block_type = OBJ;
			print_block_type();
		}

var:	VAR	{
			block_type = VAR;
			print_block_type();
		}

dat:	DAT	{
			block_type = DAT;
			print_block_type();
		}

func:
	PUB	{
			block_type = PUB;
			print_block_type();
		}
	| PRI	{
			block_type = PRI;
			print_block_type();
		}
	;


obj_decl:
	symbol ':' STRING
			{
				printf("Load object: %s -> %s\n",$1,$3);
			}
	;

var_decl:
	type var_decl	{
				printf("Variable declaration: %s -> %s\n",$1,$2);
			}
	| var_decl ',' var_decl
	| symbol
	;

type:
	BYTE_T
	| WORD_T
	| LONG_T
	;

dat_decl:
	type dat_decl	{
				printf("Dat declaration: %s -> %s\n",$1,$2);
			}
	| symbol type dat_decl
	| dat_decl ',' dat_decl
	| constantvalue
	;

function_decl:
	func WORD
			{
				printf("Function: %s\n", $2);
			}
	| func WORD parameters
			{
				printf("Function: %s\n", $2);
			}
	| func WORD parameters rvalue localvar
			{
				printf("Function: %s\n", $2);
			}
	;


rvalue:
	| ':' symbol	{
				printf("Function rvalue: %s\n",$2);
				$$ = $2;
			}
	;

parameters:
	'(' parameters ')'	{ $$ = $2; }
	| parameters ',' parameters
	| symbol	{
				printf("Symbolic parameter: %s\n", $1);
			}
	| NUMBER	{
				printf("Numeric parameter: %s\n",$1);
			}
	;

localvar:
	| '|' localvar
	| localvar ',' localvar
	| symbol {
			printf("Variable: %s\n", $1);
		}
	;

comment:
	BLOCK_COMMENT_BEGIN mltext_block BLOCK_COMMENT_CLOSE	{
						comment_type = BLOCK_COMMENT;
						$$ = $2;
						printf("Block Comment: '%s'\n",$2);
					 }
	| MLCODE_COMMENT_BEGIN mltext_block MLCODE_COMMENT_CLOSE	{
						comment_type = MLCODE_COMMENT;
						$$ = $2;
						printf("Doc ML Comment: '%s'\n",$2);
					}
	| DOC_COMMENT text_block NEWLINE		{
						comment_type = DOC_COMMENT;
						$$ = $2;
						printf("Doc Comment: '%s'\n",$2);
					}
	| CODE_COMMENT text_block NEWLINE		{
						comment_type = CODE_COMMENT;
						$$ = $2;
						printf("Code Comment: '%s'\n",$2);
					}
	;

text_block:	/* empty */
	| text_block TEXT	{ $$ = $2; }
	;

mltext_block: /* empty */
	| mltext_block TEXT		{ $$ = $2; }
	;

value:
	NUMBER
	| '$' HEX		{ $$ = $2; }
	| '%' NUMBER		{ $$ = $2; }
	| "%%" NUMBER		{ $$ = $2; }
	;

constantvalue:
	value
	| symbol
	| STRING
	| FLOAT
	;

control_statement:
	if_statement
	| repeat_statement
	| case_statement
	;

if_statement:
	IF expression			{ if_level++; printf("IF expression: %s\n",$2); }
	| IFNOT	expression		{ if_level++; printf("IFNOT expression: %s\n",$2);  }
	| ELSEIF expression		{ if_level--; printf("ELSEIF expression: %s\n",$2);  }
	| ELSEIFNOT expression		{ if_level--; printf("ELSEIFNOT expression: %s\n",$2);  }
	| ELSE				{ if_level--; printf("ELSE\n");  }
	;

case_statement:
	CASE expression			{ printf("Case expression: %s\n", $2); 			}
	| expression ':'		{ printf("Case value expression: %s\n", $1); 			}
	| OTHER ':'			{ printf("Case other\n"); 			}
	;

repeat_statement:
	REPEAT				{ printf("Repeat endless statement\n"); }
	| REPEAT expression		{ printf("Repeat expression: %s\n", $2); }
	| REPEAT symbol FROM expression TO expression step
					{ printf("Repeat %s from expression %s to expression %s step %s\n",$2,$4,$6,$7); }
	| REPEAT WHILE expression	{ printf("Repeat while expression: %s\n",$3); }
	| REPEAT UNTIL expression	{ printf("Repeat until expression: %s\n",$3); }
	| CONTINUE			{ printf("Repeat continue statement\n"); }
	| QUIT				{ printf("Repeat quit statement\n"); }
	;

step:	/* empty */
	| STEP expression		{ $$ = $2; }
	;

symbol:
	WORD				{ /* empty */ }
	;

expression:
	expression ASSIGN expression			{	printf("Variable assignment: %s = %s\n", $1, $3);		}
	| expression EQUAL expression			{	printf("Constant assignment: %s = %s\n", $1, $3);		}
	| expression BOOL_OR expression			{	printf("Logic OR: %s OR %s\n",$1,$3);				}
	| expression BOOL_OR_ASSIGN expression		{	printf("Logic OR=: %s OR= %s\n",$1,$3);				}
	| expression BOOL_AND expression		{	printf("Logic AND: %s AND %s\n",$1,$3);				}
	| expression BOOL_AND_ASSIGN expression		{	printf("Logic AND=: %s AND= %s\n",$1,$3);			}
	| expression BOOL_NOT expression		{	printf("Logic NOT: %s NOT %s\n",$1,$3);				}
	| expression BOOL_NOT_ASSIGN expression		{	printf("Logic NOT=: %s NOT= %s\n",$1,$3);			}
	| expression LT expression			{	printf("Compare LT: %s < %s\n",$1,$3);				}
	| expression GT expression			{	printf("Compare GT: %s > %s\n",$1,$3);				}
	| expression NE expression			{	printf("Compare !=: %s != %s\n",$1,$3);				}
	| expression EQUAL_EQUAL expression		{	printf("Compare Equal: %s == %s\n",$1,$3);			}
	| expression EQUAL_EQUAL_ASSIGN expression	{	printf("Compare Equal=: %s === %s\n",$1,$3);			}
	| expression EQUAL_LT expression		{	printf("Compare Equal LT: %s =< %s\n",$1,$3);			}
	| expression EQUAL_LT_ASSIGN expression		{	printf("Compare Equal LT=: %s =<= %s\n",$1,$3);			}
	| expression EQUAL_GT expression		{	printf("Compare Equal GT: %s => %s\n",$1,$3);			}
	| expression EQUAL_GT_ASSIGN expression		{	printf("Compare Equal GT=: %s =>= %s\n",$1,$3);			}
	| expression LIMIT_MIN expression		{	printf("Limit Min: %s #> %s\n",$1,$3);				}
	| expression LIMIT_MIN_ASSIGN expression	{	printf("Limit Min=: %s #>= %s\n",$1,$3);			}
	| expression LIMIT_MAX expression		{	printf("Limit Max: %s <# %s\n",$1,$3);				}
	| expression LIMIT_MAX_ASSIGN expression	{	printf("Limit Max=: %s <#= %s\n",$1,$3);			}
	| expression ADD expression			{	printf("Add: %s + %s\n",$1,$3);					}
	| expression ADD_ASSIGN expression		{	printf("Add=: %s += %s\n",$1,$3);				}
	| expression SUBTRACT expression		{	printf("Subtract: %s - %s\n",$1,$3);				}
	| expression SUBTRACT_ASSIGN expression		{	printf("Subtract=: %s -= %s\n",$1,$3);				}
	| expression MULTIPLY_LOW expression		{	printf("Multiply low: %s * %s\n",$1,$3);			}
	| expression MULTIPLY_LOW_ASSIGN expression	{	printf("Multiply low=: %s *= %s\n",$1,$3);			}
	| expression MULTIPLY_HIGH expression		{	printf("Multiply high: %s ** %s\n",$1,$3);			}
	| expression MULTIPLY_HIGH_ASSIGN expression	{	printf("Multiply high=: %s **= %s\n",$1,$3);			}
	| expression DIVIDE expression			{	printf("Divide: %s / %s\n",$1,$3);				}
	| expression DIVIDE_ASSIGN expression		{	printf("Divide=: %s /= %s\n",$1,$3);				}
	| expression MODULUS expression			{	printf("Modulus: %s // %s\n",$1,$3);				}
	| expression MODULUS_ASSIGN expression		{	printf("Modulus=: %s //= %s\n",$1,$3);				}
	| expression OR expression			{	printf("Bitwize OR: %s || %s\n",$1,$3);				}
	| expression OR_ASSIGN expression		{	printf("Bitwize OR=: %s ||= %s\n",$1,$3);			}
	| expression XOR expression			{	printf("Bitwize XOR: %s ^ %s\n",$1,$3);				}
	| expression XOR_ASSIGN expression		{	printf("Bitwize XOR=: %s ^= %s\n",$1,$3);			}
	| expression AND expression			{	printf("Bitwize AND: %s & %s\n",$1,$3);				}
	| expression AND_ASSIGN expression		{	printf("Bitwize AND=: %s &= %s\n",$1,$3);			}
	| expression ROR expression			{	printf("Rotate Right: %s -> %s\n",$1,$3);			}
	| expression ROR_ASSIGN expression		{	printf("Rotate Right=: %s ->= %s\n",$1,$3);			}
	| expression ROL expression			{	printf("Rotate Left: %s <- %s\n",$1,$3);			}
	| expression ROL_ASSIGN expression		{	printf("Rotate Left=: %s <-= %s\n",$1,$3);			}
	| expression SHR expression			{	printf("Shift Right: %s >> %s\n",$1,$3);			}
	| expression SHR_ASSIGN expression		{	printf("Shift Right=: %s >>= %s\n",$1,$3);			}
	| expression SHL expression			{	printf("Shift Left: %s << %s\n",$1,$3);				}
	| expression SHL_ASSIGN expression		{	printf("Shift Left=: %s <<= %s\n",$1,$3);			}
	| expression SAR expression			{	printf("Shift Arithmetic Right: %s ~> %s\n",$1,$3);		}
	| expression SAR_ASSIGN expression		{	printf("Shift Arithmetic Right=: %s ~>= %s\n",$1,$3);		}
	| expression REVERSE expression			{	printf("Reverse: %s >< %s\n",$1,$3);				}
	| expression REVERSE_ASSIGN expression		{	printf("Reverse=: %s ><= %s\n",$1,$3);				}
	| expression SQRT expression			{	printf("Sqrt: %s ^^ %s\n",$1,$3);				}
	| expression ABS expression			{	printf("Absolute: %s || %s\n",$1,$3);				}
	| expression DECODE expression			{	printf("Decode: %s |< %s\n",$1,$3);				}
	| expression ENCODE expression			{	printf("Encode: %s >| %s\n",$1,$3);				}
	| expression DEC				{	printf("Post Decrement: %s --\n",$1); $$ = $1;			}
	| expression INC				{	printf("Post Increment: %s ++\n",$1); $$ = $1;			}
	| expression SET 				{	printf("Post Set Assign: %s ~~\n",$1); $$ = $1;			}
	| expression RAND 				{	printf("Random Reverse Assign: %s ?\n",$1); $$ = $1;		}
	| expression CLEAR				{	printf("Post Clear: %s ~\n",$1); $$ = $1;			}
	| DECODE expression				{	printf("Decode Assign: |< %s\n",$2); $$=$2;			}
	| ENCODE expression				{	printf("Encode Assign: >| %s\n",$2); $$=$2;			}
	| DEC expression				{	printf("Pre Decrement: -- %s\n",$2); $$ = $2;			}
	| INC expression				{	printf("Pre Increment: ++ %s\n",$2); $$ = $2;			}
	| SET expression				{	printf("Sign Extend bit 15 Assign: ~~ %s\n",$2); $$ = $2;	}
	| RAND expression				{	printf("Random Forward Assign: ? %s\n",$2); $$ = $2;		}
	| CLEAR expression				{	printf("Sign Extend bit 7: ~ %s\n",$2); $$ = $2;		}
	| NOT expression				{	printf("Bitwize NOT: ! %s\n", $2); $$ = $2;			}
	| SUBTRACT expression				{	printf("Negate: - %s\n",$2); $$ = $2;				}
	| SQRT expression				{	printf("Sqrt Assign: ^^ %s\n",$2); $$=$2;			}
	| ABS expression				{	printf("Absolute Assign: || %s\n",$2); $$=$2;			}
	| AT expression					{	printf("Symbol Address: @ %s\n",$2); $$ = $2;			}
	| AT_AT expression				{	printf("Object address symbol: @@ %s\n",$2); $$ = $2;		}
	| '#' NUMBER					{	printf("CON base reference: %s\n",$2); $$ = $2;			}
	| expression '[' expression ']'			{	printf("Array expression: %s -> %s\n",$1,$3);			}
	| expression '(' expression ')'			{	printf("Parameter expression: %s -> %s\n",$1,$3);		}
	| expression '.' expression			{	printf("Object call: %s -> %s\n",$1,$3);			}
	| expression '#' expression			{	printf("Object constant reference: %s -> %s\n",$1,$3);
								char *s = malloc(strlen($1)+strlen($3)+5);
								sprintf(s,"%s#%s",$1,$3);
								$$ = s;
							}
	| expression ',' expression			{	printf("Expression list: %s , %s\n",$1,$3);			}
	| '(' expression ')'				{									}
	| symbol					{	printf("Symbol: %s\n", $1); $$ = $1;				}
	| value 					{	printf("Value: %s\n", $1); $$ = $1;				}
	| STRING 					{	printf("Literal String: %s\n", $1); $$ = $1;			}
	;

%%

int nblock_types[] = { CON, OBJ, VAR, DAT, PUB, PRI };

#ifdef STANDALONE_PARSER

void main(void)
{
	int i;

	while(i=yyparse()) {
		printf("%5d: %3d: %s -> %d\n",lineno,YY_START,yytext,i);
	}
}

#endif
