%{

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parse.tab.h"

int lineno = 1;
int last_state = 0;
int tab_count = 0;

#define save_token(x) yylval.string=strdup(x)
#define inc_lineno do { lineno++; } while(0)

void remove_underscores(char *s)
{
	char	*p, *q, *result;

	result = strdup(s);
	q = result;
	p = s;

//printf("in: '%s'\n",s);

	do {
		if (*p == '_') continue;
		*q++ = *p;
	} while (*p++);

	*q = '\0';

//printf("out: '%s'\n",result);
	yylval.string = result;
}

%}

space	[ ]+
newline	\n
number	[0-9]+[0-9_]*
hex	[0-9A-F_]+
word	[a-zA-Z_]+[a-zA-Z0-9_]*
string	\"[^\"\n]*\"
text	[^{}]+
oltext	[^\n]+
open_block_comment		"{{"
open_comment		"{"
close_comment		"}"
close_block_comment		"}}"

%s COMMENT
%s OLCOMMENT
%s HEXNUM

%%

<INITIAL>{space}	/* eat spaces */

<INITIAL>"repeat"	{ return REPEAT;		}
<INITIAL>"while"	{ return WHILE;			}
<INITIAL>"until"	{ return UNTIL;			}
<INITIAL>"from"		{ return FROM;			}
<INITIAL>"to"		{ return TO;			}
<INITIAL>"step"		{ return STEP;			}
<INITIAL>"continue"	{ return CONTINUE;		}
<INITIAL>"quit"		{ return QUIT;			}
<INITIAL>"if"		{ return IF;			}
<INITIAL>"ifnot"	{ return IFNOT;			}
<INITIAL>"else"		{ return ELSE;			}
<INITIAL>"elseif"	{ return ELSEIF;		}
<INITIAL>"elseifnot"	{ return ELSEIFNOT;		}
<INITIAL>"case"		{ return CASE;			}
<INITIAL>"other"	{ return OTHER;			}
<INITIAL>"OR="		{ return BOOL_OR_ASSIGN;	}
<INITIAL>"OR"		{ return BOOL_OR;		}
<INITIAL>"AND="		{ return BOOL_AND_ASSIGN;	}
<INITIAL>"AND"		{ return BOOL_AND;		}
<INITIAL>"NOT="		{ return BOOL_NOT_ASSIGN;	}
<INITIAL>"NOT"		{ return BOOL_NOT;		}

<INITIAL>"==="		{ return EQUAL_EQUAL_ASSIGN; }
<INITIAL>"=="		{ return EQUAL_EQUAL; }
<INITIAL>"=<="		{ return EQUAL_LT_ASSIGN; }
<INITIAL>"=<"		{ return EQUAL_LT; }
<INITIAL>"=>="		{ return EQUAL_GT_ASSIGN; }
<INITIAL>"=>"		{ return EQUAL_GT; }
<INITIAL>"="		{ return EQUAL; }
<INITIAL>":="		{ return ASSIGN; }
<INITIAL>"<#="		{ return LIMIT_MAX_ASSIGN; }
<INITIAL>"<#"		{ return LIMIT_MAX; }
<INITIAL>"<-="		{ return ROL_ASSIGN; }
<INITIAL>"<-"		{ return ROL; }
<INITIAL>"<>="		{ return NE_ASSIGN; }
<INITIAL>"<>"		{ return NE; }
<INITIAL>"<<="		{ return SHL_ASSIGN; }
<INITIAL>"<<"		{ return SHL; }
<INITIAL>"<"		{ return LT; }
<INITIAL>"><="		{ return REVERSE_ASSIGN; }
<INITIAL>"><"		{ return REVERSE; }
<INITIAL>">|"		{ return ENCODE; }
<INITIAL>">>="		{ return SHR_ASSIGN; }
<INITIAL>">>"		{ return SHR; }
<INITIAL>">"		{ return GT; }
<INITIAL>"++"		{ return INC; }
<INITIAL>"+="		{ return ADD_ASSIGN; }
<INITIAL>"+"		{ return ADD; }
<INITIAL>"->="		{ return ROR_ASSIGN; }
<INITIAL>"->"		{ return ROR; }
<INITIAL>"--"		{ return DEC; }
<INITIAL>"-="		{ return SUBTRACT_ASSIGN; }
<INITIAL>"-"		{ return SUBTRACT; }
<INITIAL>"**="		{ return MULTIPLY_HIGH_ASSIGN; }
<INITIAL>"**"		{ return MULTIPLY_HIGH; }
<INITIAL>"*="		{ return MULTIPLY_LOW_ASSIGN; }
<INITIAL>"*"		{ return MULTIPLY_LOW; }
<INITIAL>"//="		{ return MODULUS_ASSIGN; }
<INITIAL>"//"		{ return MODULUS; }
<INITIAL>"/="		{ return DIVIDE_ASSIGN; }
<INITIAL>"/"		{ return DIVIDE; }
<INITIAL>"|<"		{ return DECODE; }
<INITIAL>"|="		{ return OR_ASSIGN; }
<INITIAL>"|"		{ return OR; }
<INITIAL>"^^"		{ return SQRT; }
<INITIAL>"^="		{ return XOR_ASSIGN; }
<INITIAL>"^"		{ return XOR; }
<INITIAL>"&="		{ return AND_ASSIGN; }
<INITIAL>"&"		{ return AND; }
<INITIAL>"!"		{ return NOT; }
<INITIAL>"~>="		{ return SAR_ASSIGN; }
<INITIAL>"~>"		{ return SAR; }
<INITIAL>"~~"		{ return SET; }
<INITIAL>"~"		{ return CLEAR; }
<INITIAL>"#>="		{ return LIMIT_MIN_ASSIGN; }
<INITIAL>"#>"		{ return LIMIT_MIN; }

<INITIAL>"?"		{ return RAND; }
<INITIAL>"@@"		{ return AT_AT; }
<INITIAL>"@"		{ return AT; }
<INITIAL>\t		{ tab_count++; }
<INITIAL>"byte"		{ save_token(yytext); return BYTE_T; }
<INITIAL>"word"		{ save_token(yytext); return WORD_T; }
<INITIAL>"long"		{ save_token(yytext); return LONG_T; }

"CON"			{ return CON; }

"OBJ"			{ return OBJ; }

"VAR"			{ return VAR; }

"PUB"			{ return PUB; }

"PRI"			{ return PRI; }

<INITIAL>{word}		{ save_token(yytext); return WORD; }

{string}	{ 
			save_token(yytext+1); 
			*(strchr(yylval.string,'"'))='\0';
			return STRING; 
		}

<INITIAL>{newline}	{ inc_lineno; tab_count = 0; }

{number}	{ remove_underscores(yytext); return NUMBER; }

<INITIAL>\$	{ BEGIN HEXNUM; return yytext[0]; }

<HEXNUM>{hex}	{ 
			remove_underscores(yytext); 
			BEGIN INITIAL; 
			return HEX;
		}

<INITIAL>{open_block_comment} {
			BEGIN COMMENT; return BLOCK_COMMENT_BEGIN;
		}

<INITIAL>{open_comment} {
			BEGIN COMMENT; return MLCODE_COMMENT_BEGIN;
		}

<COMMENT>{newline}	{ inc_lineno; }

<COMMENT>{open_comment} {
			return yytext[0];
		}

<COMMENT>{text} {
			 save_token(yytext); return TEXT;
		}

<INITIAL>"''" {
			BEGIN OLCOMMENT; return DOC_COMMENT;
		}

<INITIAL>"'" {
			BEGIN OLCOMMENT; return CODE_COMMENT;
		}

<OLCOMMENT>{oltext} {
			 save_token(yytext); return TEXT;
		}

<OLCOMMENT>{newline} {
			save_token(yytext);
			BEGIN INITIAL;
			inc_lineno;
			return NEWLINE;
		}

<COMMENT>{close_block_comment} {
			BEGIN INITIAL; return BLOCK_COMMENT_CLOSE;
		}

<COMMENT>{close_comment} {
			BEGIN INITIAL; return MLCODE_COMMENT_CLOSE;
		}

.		{ return yytext[0]; }

%%

int yywrap(void)
{
	return 1;
}

#ifdef STANDALONE

YYSTYPE yylval;

void main(void)
{
	int	i;
	int	cnt=0;

	while(i=yylex()) {
		cnt++;
		printf("%5d:%5d:%5d:\t%s -> %d\n",lineno,cnt,YY_START,yytext,i);
	}
}
#endif
